/*
        Astronomy FOCUS Focuser
    Copyright (c) 2022 Vittorio Signorelli

        Based on Moonlite Focuser
    Copyright (C) 2013-2019 Jasem Mutlaq (mutlaqja@ikarustech.com)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "astrofocus.h"

#include "indicom.h"

#include <cmath>
#include <cstring>
#include <memory>

#include <termios.h>
#include <unistd.h>

static std::unique_ptr<AstronomyFOCUS> astronomyFOCUS(new AstronomyFOCUS());

AstronomyFOCUS::AstronomyFOCUS()
{
    setVersion(1, 0);

    // Can move in Absolute & Relative motions, can AbortFocuser motion, and has variable speed.
    FI::SetCapability(FOCUSER_CAN_ABS_MOVE | FOCUSER_CAN_ABORT | FOCUSER_CAN_REVERSE);
    setSupportedConnections(CONNECTION_SERIAL);

}

bool AstronomyFOCUS::initProperties()
{
    INDI::Focuser::initProperties();

    // Focuser temperature
    IUFillNumber(&TemperatureN[0], "TEMPERATURE", "Celsius", "%6.2f", -50, 70., 0., 0.);
    IUFillNumberVector(&TemperatureNP, TemperatureN, 1, getDeviceName(), "FOCUS_TEMPERATURE", "Temperature",
                       MAIN_CONTROL_TAB, IP_RO, 0, IPS_IDLE);

    /* A movement */

    FocusAbsPosN[0].min   = 0.;
    FocusAbsPosN[0].max   = 2000000.;
    FocusAbsPosN[0].value = 0;
    FocusAbsPosN[0].step  = 1;

    setPollingPeriodRange(1000, 30000);
    setDefaultPollingPeriod(1000);

    return true;
}

bool AstronomyFOCUS::updateProperties()
{
    INDI::Focuser::updateProperties();

    if (isConnected())
    {
        defineProperty(&TemperatureNP);
        GetFocusParams();

        LOG_INFO("AstronomyFOCUS parameters updated, focuser ready for use.");
    }
    else
    {
        deleteProperty(TemperatureNP.name);
    }

    return true;
}

bool AstronomyFOCUS::Handshake()
{
    if (Ack())
    {
        LOG_INFO("AstronomyFOCUS is online. Getting focus parameters...");
        readReverseDirection();
        return true;
    }

    LOG_INFO(
        "Error retrieving data from AstronomyFOCUS, please ensure AstronomyFOCUS controller is powered and the port is correct.");
    return false;
}

const char * AstronomyFOCUS::getDefaultName()
{
    return "AstronomyFOCUS";
}

bool AstronomyFOCUS::Ack()
{
    bool success = false;

    for (int i = 0; i < 3; i++)
    {
        if (readVersion())
        {
            success = true;
            break;
        }

        sleep(1);
    }

    return success;
}


bool AstronomyFOCUS::readVersion()
{
    char res[ML_RES] = {0};

    if (sendCommand("GIN#", res, true, 2) == false)
        return false;

    LOGF_INFO("Detected firmware version %c.%c.%c", res[0], res[2], res[4]);

    return true;
}

bool AstronomyFOCUS::readReverseDirection()
{
    char res[ML_RES] = {0};

    if (sendCommand("FGD#", res) == false)
        return false;

    uint32_t temp = 0;

    int rc = sscanf(res, "%u#", &temp);

    if (rc > 0)
    {
        if(temp == 0)
        {
            FocusReverseS[INDI_DISABLED].s = ISS_ON;
        }
        else if (temp == 1)
        {
            FocusReverseS[INDI_ENABLED].s = ISS_ON;
        }
        else
        {
            LOGF_ERROR("Invalid Response: focuser Reverse direction value (%s)", res);
            return false;
        }
    }
    else
    {
        LOGF_ERROR("Unknown error: focuser Reverse direction value (%s)", res);
        return false;
    }
    return true;
}

bool AstronomyFOCUS::readTemperature()
{
    char res[ML_RES] = {0};

    if (sendCommand("FGT#", res) == false)
        return false;

    double temp = 0;
    int rc = sscanf(res, "%lf#", &temp);
    if (rc > 0)
    {
        TemperatureN[0].value = temp;
    }
    else
    {
        LOGF_ERROR("Unknown error: focuser temperature value (%s)", res);
        return false;
    }
    return true;
}

bool AstronomyFOCUS::readPosition()
{
    char res[ML_RES] = {0};

    if (sendCommand("FGP#", res) == false)
        return false;

    int32_t pos;
    int rc = sscanf(res, "%d#", &pos);

    if (rc > 0)
        FocusAbsPosN[0].value = pos;
    else
    {
        LOGF_ERROR("Unknown error: focuser position value (%s)", res);
        return false;
    }

    return true;
}


bool AstronomyFOCUS::isMoving()
{
    char res[ML_RES] = {0};

    if (sendCommand("FGM#", res) == false)
        return false;

    if (strstr(res, "M#"))
        return true;
    else if (strstr(res, "S#"))
        return false;

    LOGF_ERROR("Unknown error: isMoving value (%s)", res);
    return false;
}


bool AstronomyFOCUS::MoveFocuser(uint32_t position)
{
    char cmd[ML_RES] = {0};
    snprintf(cmd, ML_RES, "FSA%d#", position);
    if (sendCommand(cmd) == false)
        return false;

    return true;
}


void AstronomyFOCUS::GetFocusParams()
{
    if (readPosition())
        IDSetNumber(&FocusAbsPosNP, nullptr);

    if (readTemperature())
        IDSetNumber(&TemperatureNP, nullptr);
}

bool AstronomyFOCUS::ReverseFocuser(bool enable)
{
    char cmd[ML_RES] = {0};
    if (enable) LOG_INFO("Reverse Motor direction enabled");
    else LOG_INFO("Reverse Motor direction disabled");
    snprintf(cmd, ML_RES, "FSD%d#", static_cast<int>(enable));
    return sendCommand(cmd);
}


void AstronomyFOCUS::timedMoveHelper(void * context)
{
    static_cast<AstronomyFOCUS *>(context)->timedMoveCallback();
}

void AstronomyFOCUS::timedMoveCallback()
{
    AbortFocuser();
    FocusAbsPosNP.s = IPS_IDLE;
    FocusTimerNP.s = IPS_IDLE;
    FocusTimerN[0].value = 0;
    IDSetNumber(&FocusAbsPosNP, nullptr);
    IDSetNumber(&FocusTimerNP, nullptr);
}

IPState AstronomyFOCUS::MoveAbsFocuser(uint32_t targetTicks)
{
    targetPos = targetTicks;

    if (!MoveFocuser(targetPos))
        return IPS_ALERT;

    return IPS_BUSY;
}


void AstronomyFOCUS::TimerHit()
{
    if (!isConnected())
        return;

    bool rc = readPosition();
    if (rc)
    {
        if (fabs(lastPos - FocusAbsPosN[0].value) > 5)
        {
            IDSetNumber(&FocusAbsPosNP, nullptr);
            lastPos = static_cast<uint32_t>(FocusAbsPosN[0].value);
        }
    }

    rc = readTemperature();
    if (rc)
    {
        if (fabs(lastTemperature - TemperatureN[0].value) >= 0.5)
        {
            IDSetNumber(&TemperatureNP, nullptr);
            lastTemperature = static_cast<uint32_t>(TemperatureN[0].value);
        }
    }

    if (FocusAbsPosNP.s == IPS_BUSY || FocusRelPosNP.s == IPS_BUSY)
    {
        if (!isMoving())
        {
            FocusAbsPosNP.s = IPS_OK;
            IDSetNumber(&FocusAbsPosNP, nullptr);
            lastPos = static_cast<uint32_t>(FocusAbsPosN[0].value);
            LOG_INFO("Focuser reached requested position.");
        }
    }

    SetTimer(getCurrentPollingPeriod());
}

bool AstronomyFOCUS::AbortFocuser()
{
    LOG_INFO("Motor Stopped.");
    return sendCommand("FSH#");
}


bool AstronomyFOCUS::saveConfigItems(FILE * fp)
{
    Focuser::saveConfigItems(fp);


    return true;
}

bool AstronomyFOCUS::sendCommand(const char * cmd, char * res, bool silent, int nret)
{
    int nbytes_written = 0, nbytes_read = 0, rc = -1;

    tcflush(PortFD, TCIOFLUSH);

    LOGF_DEBUG("CMD <%s>", cmd);

    if ((rc = tty_write_string(PortFD, cmd, &nbytes_written)) != TTY_OK)
    {
        char errstr[MAXRBUF] = {0};
        tty_error_msg(rc, errstr, MAXRBUF);
        if (!silent)
            LOGF_ERROR("Serial write error: %s.", errstr);
        return false;
    }

    if (res == nullptr)
    {
        tcdrain(PortFD);
        return true;
    }

    // this is to handle the GV command which doesn't return the terminator, use the number of chars expected
    if (nret == 0)
    {
        rc = tty_nread_section(PortFD, res, ML_RES, ML_DEL, ML_TIMEOUT, &nbytes_read);
    }
    else
    {
        rc = tty_read(PortFD, res, nret, ML_TIMEOUT, &nbytes_read);
    }
    if (rc != TTY_OK)
    {
        char errstr[MAXRBUF] = {0};
        tty_error_msg(rc, errstr, MAXRBUF);
        if (!silent)
            LOGF_ERROR("Serial read error: %s.", errstr);
        return false;
    }

    LOGF_DEBUG("RES <%s>", res);

    tcflush(PortFD, TCIOFLUSH);

    return true;
}
